import abc
import typing
import inspect
import dataclasses

import sklearn.base

class FeatureExtractorBase(sklearn.base.TransformerMixin, abc.ABC):
    @dataclasses.dataclass(frozen = True)
    class Presets:
        full: object
        reduced: object
        quick: object

    PRESETS: Presets = None

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        if 'settings' not in inspect.signature(cls.__init__).parameters.keys():
            return
        if cls.PRESETS is None:
            raise TypeError(f"{cls.__name__} must define PRESETS")
        if not isinstance(cls.PRESETS, FeatureExtractorBase.Presets):
            raise TypeError(f"{cls.__name__} must be an instance of Presets")


    def nest_feature_names(self, prefix: str, names: typing.List[str]) -> typing.List[str]:
        return [ f"{prefix} -> {name}" for name in names ]

    @abc.abstractmethod
    def feature_names(self) -> typing.List[str]:
        pass



import torch


if torch.cuda.is_available():
    default_device = torch.device('cuda')
elif torch.mps.is_available():
    default_device = torch.device('mps')
else:
    default_device = torch.device('cpu')



import torch

class Dataset:
    def __init__(self, images, features, masks):
        self.images = torch.tensor(images, dtype = torch.float32)
        if len(self.images.shape) == 4:
            if self.images.shape[-1] == 3:
                self.images = self.images.permute((0, 3, 1, 2))
        else:
            assert len(self.images.shape) == 3
            self.images = self.images.unsqueeze(1)

        self.masks = torch.tensor(masks, dtype = torch.float32).unsqueeze(1)
        self.features = torch.tensor(features, dtype = torch.float32)



import typing
import dataclasses

import numpy
import skimage.measure
import skimage.feature

import cvtda.utils
import cvtda.logging

from . import utils
from .interface import Extractor


class GrayGeometryExtractor(cvtda.utils.FeatureExtractorBase):
    @dataclasses.dataclass(frozen = True)
    class Settings:
        reduced_stats: bool = True

        daisy: bool = True
        sift: bool = True
        orb: bool = True
        hog: bool = True
        basic: bool = True
        blur_effect: bool = True
        centroid: bool = True
        inertia_tensor_eigvals: bool = True
        moments: bool = True
        moments_central: bool = True
        moments_hu: bool = True
        shannon_entropy: bool = True
        curvature: bool = True

    PRESETS = cvtda.utils.FeatureExtractorBase.Presets(
        full = Settings(reduced_stats = False), reduced = Settings(), quick = Settings(curvature = False)
    )

    FEATURE_NAMES = [
        "daisy", "sift", "orb", "hog", "basic", "blur_effect", "centroid", "inertia_tensor_eigvals",
        "moments", "moments_central", "moments_hu", "shannon_entropy", "curvature"
    ]
    
    def __init__(self, n_jobs: int = -1, settings = Settings()):
        self.fitted_ = False
        self.n_jobs_ = n_jobs
        self.feature_names_ = []
        self.reduced_stats_ = settings.reduced_stats
        self.features_mask = [
            settings.daisy, settings.sift, settings.orb, settings.hog, settings.basic,
            settings.blur_effect, settings.centroid, settings.inertia_tensor_eigvals,
            settings.moments, settings.moments_central, settings.moments_hu,
            settings.shannon_entropy, settings.curvature
        ]
        

    def feature_names(self) -> typing.List[str]:
        assert self.fitted_ is True, 'fit() must be called before feature_names()'
        return self.feature_names_

    def fit(self, gray_images: numpy.ndarray):
        self.feature_names_ = []
        for is_needed, calc, name in zip(self.features_mask, GrayGeometryExtractor.FEATURE_CALCULATORS, GrayGeometryExtractor.FEATURE_NAMES):
            if is_needed:
                self.feature_names_.extend([ f"{name}-{i}" for i in range(len(calc(self, gray_images[0]))) ])
        self.fitted_ = True
        return self

    def transform(self, gray_images: numpy.ndarray) -> numpy.ndarray:
        assert self.fitted_ is True, 'fit() must be called before transform()'
        
        def process_one_(gray_image: numpy.ndarray) -> numpy.ndarray:
            return numpy.nan_to_num(numpy.concatenate(self.calc_raw_(gray_image)), 0)
        pbar = cvtda.logging.logger().pbar(gray_images, desc = 'GrayGeometryExtractor')
        features = numpy.stack(cvtda.utils.parallel(process_one_, pbar, n_jobs = self.n_jobs_))
        assert features.shape == (len(gray_images), len(self.feature_names()))
        return features

    def calc_raw_(self, gray_image: numpy.ndarray) -> typing.List[numpy.ndarray]:
        return [
            calc(self, gray_image)
            for is_needed, calc in zip(self.features_mask, GrayGeometryExtractor.FEATURE_CALCULATORS)
            if is_needed
        ]
    
    def calc_daisy_(self, gray_image: numpy.ndarray) -> numpy.ndarray:
        image_shape = max(*gray_image.shape)
        return skimage.feature.daisy(
            gray_image,
            step = (6 * image_shape // 32),
            radius = (12 * image_shape // 32),
            rings = 5,
            histograms = 5,
            orientations = 8
        ).flatten()

    def calc_sift_(self, gray_image: numpy.ndarray) -> numpy.ndarray:
        sift = skimage.feature.SIFT()
        try:
            sift.detect_and_extract(gray_image)
            sift_descriptors = sift.descriptors.transpose()
            if sift_descriptors.shape[1] == 0:
                raise 'How is this possible?'
        except:
            sift_descriptors = numpy.zeros((128,1))
        return cvtda.utils.sequence2features(numpy.ma.array(sift_descriptors), reduced = self.reduced_stats_).flatten()

    def calc_orb_(self, gray_image: numpy.ndarray) -> numpy.ndarray:
        orb = skimage.feature.ORB()
        try:
            orb.detect_and_extract(gray_image)
            orb_descriptors = orb.descriptors.transpose()
            if orb_descriptors.shape[1] == 0:
                raise 'How is this possible?'
        except:
            orb_descriptors = numpy.zeros((256,1))
        return cvtda.utils.sequence2features(numpy.ma.array(orb_descriptors), reduced = self.reduced_stats_).flatten()

    def calc_hog_(self, gray_image: numpy.ndarray) -> numpy.ndarray:
        return skimage.feature.hog(gray_image, pixels_per_cell = (gray_image.shape[0] // 4, gray_image.shape[1] // 4))
    
    def calc_basic_(self, gray_image: numpy.ndarray) -> numpy.ndarray:
        basic_features = skimage.feature.multiscale_basic_features(gray_image).reshape((-1, 24))
        return cvtda.utils.sequence2features(
            numpy.ma.array(basic_features.transpose()), reduced = self.reduced_stats_
        ).flatten()
            
    def calc_blur_effect_(self, gray_image: numpy.ndarray) -> numpy.ndarray:
        return numpy.array([ skimage.measure.blur_effect(gray_image) ])
    
    def calc_centroid_(self, gray_image: numpy.ndarray) -> numpy.ndarray:
        return skimage.measure.centroid(gray_image)
    
    def calc_inertia_tensor_eigvals_(self, gray_image: numpy.ndarray) -> numpy.ndarray:
        try:
            return numpy.array(skimage.measure.inertia_tensor_eigvals(gray_image))
        except:
            return numpy.zeros((2))
    
    def calc_moments_(self, gray_image: numpy.ndarray) -> numpy.ndarray:
        return skimage.measure.moments(gray_image, order = 9).flatten()
    
    def calc_moments_central_(self, gray_image: numpy.ndarray) -> numpy.ndarray:
        return skimage.measure.moments_central(gray_image, order = 9).flatten()
    
    def calc_moments_hu_(self, gray_image: numpy.ndarray) -> numpy.ndarray:
        moments_central = skimage.measure.moments_central(gray_image, order = 9)
        moments_normalized = skimage.measure.moments_normalized(moments_central)
        return skimage.measure.moments_hu(moments_normalized).flatten()
    
    def calc_shannon_entropy_(self, gray_image: numpy.ndarray):
        return numpy.array([ skimage.measure.shannon_entropy(gray_image) ])
        
    def calc_curvature_(self, gray_image: numpy.ndarray) -> numpy.ndarray:
        min_, max_= gray_image.min(), gray_image.max()
        assert (min_ >= 0) and (max_ <= 1), f'Bad image format: should be [0, 1]; received [{min_}, {max_}]'

        euler_numbers, area, perimeter = [], [], []
        for threshold in range(256):
            bin = (gray_image > threshold / 255.)
            euler_numbers.append(skimage.measure.euler_number(bin))
            area.append(bin.sum())
            perimeter.append(skimage.measure.perimeter(bin))
        series = numpy.array([ euler_numbers, area, perimeter ])
        series_diff = numpy.array([ numpy.diff(euler_numbers), numpy.diff(area), numpy.diff(perimeter) ])

        return numpy.concatenate([
            cvtda.utils.sequence2features(series, reduced = self.reduced_stats_).flatten(),
            cvtda.utils.sequence2features(series_diff, reduced = self.reduced_stats_).flatten(),
        ])

    FEATURE_CALCULATORS = [
        calc_daisy_, calc_sift_, calc_orb_, calc_hog_, calc_basic_,
        calc_blur_effect_, calc_centroid_, calc_inertia_tensor_eigvals_,
        calc_moments_, calc_moments_central_, calc_moments_hu_,
        calc_shannon_entropy_, calc_curvature_
    ]


class RGBGeometryExtractor(cvtda.utils.FeatureExtractorBase):
    @dataclasses.dataclass(frozen = True)
    class Settings:
        reduced_stats: bool = True

        hog: bool = True
        centroid: bool = True
        inertia_tensor_eigvals: bool = True
        moments: bool = True
        moments_central: bool = True
        corr_coef: bool = True

    PRESETS = cvtda.utils.FeatureExtractorBase.Presets(
        full = Settings(reduced_stats = False), reduced = Settings(), quick = Settings()
    )

    FEATURE_NAMES = [ "hog", "centroid", "inertia_tensor_eigvals", "moments", "moments_central", "corr_coef" ]
    
    def __init__(self, n_jobs: int = -1, settings = Settings()):
        self.fitted_ = False
        self.n_jobs_ = n_jobs
        self.feature_names_ = []
        self.reduced_stats_ = settings.reduced_stats
        self.features_mask = [
            settings.hog, settings.centroid, settings.inertia_tensor_eigvals,
            settings.moments, settings.moments_central, settings.corr_coef
        ]

    def feature_names(self) -> typing.List[str]:
        assert self.fitted_ is True, 'fit() must be called before feature_names()'
        return self.feature_names_

    def fit(self, rgb_images: numpy.ndarray):
        self.feature_names_ = []
        for is_needed, calc, name in zip(self.features_mask, RGBGeometryExtractor.FEATURE_CALCULATORS, RGBGeometryExtractor.FEATURE_NAMES):
            if is_needed:
                self.feature_names_.extend([ f"{name}-{i}" for i in range(len(calc(self, rgb_images[0]))) ])
        self.fitted_ = True
        return self

    def transform(self, rgb_images: numpy.ndarray) -> numpy.ndarray:
        assert self.fitted_ is True, 'fit() must be called before transform()'

        def process_one_(rgb_image: numpy.ndarray) -> numpy.ndarray:
            return numpy.nan_to_num(numpy.concatenate(self.calc_raw_(rgb_image)), 0)
        pbar = cvtda.logging.logger().pbar(rgb_images, desc = 'RGBGeometryExtractor')
        features = numpy.stack(cvtda.utils.parallel(process_one_, pbar, n_jobs = self.n_jobs_))
        assert features.shape == (len(rgb_images), len(self.feature_names()))
        return features
    
    def calc_raw_(self, rgb_image: numpy.ndarray) -> typing.List[numpy.ndarray]:
        return [
            calc(self, rgb_image)
            for is_needed, calc in zip(self.features_mask, RGBGeometryExtractor.FEATURE_CALCULATORS)
            if is_needed
        ]
    
    def calc_hog_(self, rgb_image: numpy.ndarray) -> numpy.ndarray:
        return skimage.feature.hog(rgb_image, channel_axis = 2)
    
    def calc_centroid_(self, rgb_image: numpy.ndarray) -> numpy.ndarray:
        return skimage.measure.centroid(rgb_image)
    
    def calc_inertia_tensor_eigvals_(self, rgb_image: numpy.ndarray) -> numpy.ndarray:
        try:
            return numpy.array(skimage.measure.inertia_tensor_eigvals(rgb_image))
        except:
            return numpy.zeros((3))
    
    def calc_moments_(self, rgb_image: numpy.ndarray) -> numpy.ndarray:
        return skimage.measure.moments(rgb_image, order = 6).flatten()
    
    def calc_moments_central_(self, rgb_image: numpy.ndarray) -> numpy.ndarray:
        return skimage.measure.moments_central(rgb_image, order = 6).flatten()
    
    def calc_corr_coef_(self, rgb_image: numpy.ndarray) -> numpy.ndarray:
        return numpy.array([
            skimage.measure.pearson_corr_coeff(rgb_image[:, :, 0], rgb_image[:, :, 1])[0],
            skimage.measure.pearson_corr_coeff(rgb_image[:, :, 0], rgb_image[:, :, 2])[0],
            skimage.measure.pearson_corr_coeff(rgb_image[:, :, 1], rgb_image[:, :, 2])[0],
        ])
    
    FEATURE_CALCULATORS = [
        calc_hog_, calc_centroid_, calc_inertia_tensor_eigvals_,
        calc_moments_, calc_moments_central_, calc_corr_coef_
    ]


class MultidimensionalGeometryExtractor(cvtda.utils.FeatureExtractorBase):
    @dataclasses.dataclass(frozen = True)
    class Settings:
        reduced_stats: bool = True

        basic: bool = True
        blur_effect: bool = True
        centroid: bool = True
        inertia_tensor_eigvals: bool = True
        moments: bool = True
        moments_central: bool = True

    PRESETS = cvtda.utils.FeatureExtractorBase.Presets(
        full = Settings(reduced_stats = False), reduced = Settings(), quick = Settings()
    )

    FEATURE_NAMES = [ "basic", "blur_effect", "centroid", "inertia_tensor_eigvals", "moments", "moments_central" ]
    
    def __init__(self, n_jobs: int = -1, settings = Settings()):
        self.fitted_ = False
        self.n_jobs_ = n_jobs
        self.feature_names_ = []
        self.reduced_stats_ = settings.reduced_stats
        self.features_mask = [
            settings.basic, settings.blur_effect, settings.centroid,
            settings.inertia_tensor_eigvals, settings.moments, settings.moments_central
        ]
    
    def feature_names(self) -> typing.List[str]:
        assert self.fitted_ is True, 'fit() must be called before feature_names()'
        return self.feature_names_

    def fit(self, nd_images: numpy.ndarray):
        self.feature_names_ = []
        for is_needed, calc, name in zip(self.features_mask, MultidimensionalGeometryExtractor.FEATURE_CALCULATORS, MultidimensionalGeometryExtractor.FEATURE_NAMES):
            if is_needed:
                self.feature_names_.extend([ f"{name}-{i}" for i in range(len(calc(self, nd_images[0]))) ])
        self.fitted_ = True
        return self

    def transform(self, nd_images: numpy.ndarray) -> numpy.ndarray:
        assert self.fitted_ is True, 'fit() must be called before transform()'
        
        def process_one_(nd_image: numpy.ndarray) -> numpy.ndarray:
            return numpy.nan_to_num(numpy.concatenate(self.calc_raw_(nd_image)), 0)
        pbar = cvtda.logging.logger().pbar(nd_images, desc = 'MultidimensionalGeometryExtractor')
        features = numpy.stack(cvtda.utils.parallel(process_one_, pbar, n_jobs = self.n_jobs_))
        assert features.shape == (len(nd_images), len(self.feature_names()))
        return features
    
    def calc_raw_(self, nd_image: numpy.ndarray) -> typing.List[numpy.ndarray]:
        return [
            calc(self, nd_image)
            for is_needed, calc in zip(self.features_mask, MultidimensionalGeometryExtractor.FEATURE_CALCULATORS)
            if is_needed
        ]
    
    def calc_basic_(self, nd_image: numpy.ndarray) -> typing.Optional[numpy.ndarray]:
        basic_features = skimage.feature.multiscale_basic_features(nd_image)
        basic_features = basic_features.reshape((-1, basic_features.shape[-1]))
        return cvtda.utils.sequence2features(
            numpy.ma.array(basic_features.transpose()), reduced = self.reduced_stats_
        ).flatten()
    
    def calc_blur_effect_(self, nd_image: numpy.ndarray) -> typing.Optional[numpy.ndarray]:
        return numpy.array([ skimage.measure.blur_effect(nd_image) ])
    
    def calc_centroid_(self, nd_image: numpy.ndarray) -> typing.Optional[numpy.ndarray]:
        return skimage.measure.centroid(nd_image)
    
    def calc_inertia_tensor_eigvals_(self, nd_image: numpy.ndarray) -> typing.Optional[numpy.ndarray]:
        try:
            return numpy.array(skimage.measure.inertia_tensor_eigvals(nd_image))
        except:
            return numpy.zeros((len(nd_image.shape)))
    
    def calc_moments_(self, nd_image: numpy.ndarray) -> typing.Optional[numpy.ndarray]:
        return skimage.measure.moments(nd_image, order = 9).flatten()
    
    def calc_moments_central_(self, nd_image: numpy.ndarray) -> typing.Optional[numpy.ndarray]:
        return skimage.measure.moments_central(nd_image, order = 9).flatten()
    
    FEATURE_CALCULATORS = [
        calc_basic_, calc_blur_effect_, calc_centroid_,
        calc_inertia_tensor_eigvals_, calc_moments_, calc_moments_central_
    ]

class GeometryExtractor(Extractor):
    @dataclasses.dataclass(frozen = True)
    class Settings:
        rgb: RGBGeometryExtractor.Settings = RGBGeometryExtractor.Settings()
        gray: GrayGeometryExtractor.Settings = GrayGeometryExtractor.Settings()
        multidimensional: MultidimensionalGeometryExtractor.Settings = MultidimensionalGeometryExtractor.Settings()

    PRESETS = cvtda.utils.FeatureExtractorBase.Presets(
        full = Settings(
            rgb = RGBGeometryExtractor.PRESETS.full,
            gray = GrayGeometryExtractor.PRESETS.full,
            multidimensional = MultidimensionalGeometryExtractor.PRESETS.full
        ),
        reduced = Settings(
            rgb = RGBGeometryExtractor.PRESETS.reduced,
            gray = GrayGeometryExtractor.PRESETS.reduced,
            multidimensional = MultidimensionalGeometryExtractor.PRESETS.reduced
        ),
        quick = Settings(
            rgb = RGBGeometryExtractor.PRESETS.quick,
            gray = GrayGeometryExtractor.PRESETS.quick,
            multidimensional = MultidimensionalGeometryExtractor.PRESETS.quick
        )
    )

    def __init__(self, n_jobs: int = -1, settings = Settings(), only_get_from_dump: bool = False):
        super().__init__(n_jobs = n_jobs, settings = settings, only_get_from_dump = only_get_from_dump)

        self.rgb_extractor_ = RGBGeometryExtractor(n_jobs = self.n_jobs_, settings = settings.rgb)
        self.gray_extractor_ = GrayGeometryExtractor(n_jobs = self.n_jobs_, settings = settings.gray)
        self.multidimensional_extractor_ = MultidimensionalGeometryExtractor(n_jobs = self.n_jobs_, settings = settings.multidimensional)


    def process_rgb_(self, rgb_images: numpy.ndarray, do_fit: bool, dump_name: typing.Optional[str] = None) -> numpy.ndarray:
        return utils.process_iter_dump(self.rgb_extractor_, rgb_images, do_fit, self.features_dump_(dump_name))
    
    def feature_names_rgb_(self) -> typing.List[str]:
        return self.rgb_extractor_.feature_names()

    def process_gray_(self, gray_images: numpy.ndarray, do_fit: bool, dump_name: typing.Optional[str] = None) -> numpy.ndarray:
        if len(gray_images.shape) == 3:
            return utils.process_iter_dump(self.gray_extractor_, gray_images, do_fit, self.features_dump_(dump_name))
        else:
            return utils.process_iter_dump(self.multidimensional_extractor_, gray_images, do_fit, self.features_dump_(dump_name))

    def feature_names_gray_(self) -> typing.List[str]:
        if len(self.fit_dimensions_) == 2:
            return self.gray_extractor_.feature_names()
        else:
            return self.multidimensional_extractor_.feature_names()



import typing
import dataclasses

import numpy
import gtda.images
import sklearn.preprocessing

import cvtda.utils
import cvtda.logging
import cvtda.dumping

from . import utils
from .FiltrationsExtractor import FiltrationsExtractor
from .GreyscaleExtractor import GreyscaleExtractor
from .PointCloudsExtractor import PointCloudsExtractor
from .GeometryExtractor import GeometryExtractor


class FeatureExtractor(cvtda.utils.FeatureExtractorBase):
    @dataclasses.dataclass(frozen = True)
    class Settings:
        greyscale: GreyscaleExtractor.Settings = GreyscaleExtractor.Settings()
        inverted: GreyscaleExtractor.Settings = GreyscaleExtractor.Settings()
        filtrations: FiltrationsExtractor.Settings = FiltrationsExtractor.Settings()
        point_clouds: PointCloudsExtractor.Settings = PointCloudsExtractor.Settings()
        geometry: GeometryExtractor.Settings = GeometryExtractor.Settings()

    PRESETS = cvtda.utils.FeatureExtractorBase.Presets(
        full = Settings(
            greyscale=GreyscaleExtractor.PRESETS.full,
            inverted=GreyscaleExtractor.PRESETS.full,
            filtrations=FiltrationsExtractor.PRESETS.full,
            point_clouds=PointCloudsExtractor.PRESETS.full,
            geometry=GeometryExtractor.PRESETS.full,
        ),
        reduced = Settings(
            greyscale=GreyscaleExtractor.PRESETS.reduced,
            inverted=GreyscaleExtractor.PRESETS.reduced,
            filtrations=FiltrationsExtractor.PRESETS.reduced,
            point_clouds=PointCloudsExtractor.PRESETS.reduced,
            geometry=GeometryExtractor.PRESETS.reduced,
        ),
        quick = Settings(
            greyscale=GreyscaleExtractor.PRESETS.quick,
            inverted=GreyscaleExtractor.Settings(enabled = False),
            filtrations=FiltrationsExtractor.PRESETS.quick,
            point_clouds=PointCloudsExtractor.PRESETS.quick,
            geometry=GeometryExtractor.PRESETS.quick,
        ),
    )

    def __init__(
        self,
        n_jobs: int = -1,
        return_diagrams: bool = False,
        only_get_from_dump: bool = False,
        settings: Settings = PRESETS.reduced,
    ):
        self.fitted_ = False
        self.return_diagrams_ = return_diagrams
        self.only_get_from_dump_ = only_get_from_dump
        
        extractor_kwargs = { 'n_jobs': n_jobs, 'only_get_from_dump': only_get_from_dump }
        topological_extractor_kwargs = { **extractor_kwargs, 'return_diagrams': return_diagrams }

        self.inverter_ = gtda.images.Inverter()
        self.greyscale_ = GreyscaleExtractor(settings = settings.greyscale, **topological_extractor_kwargs)
        self.inverted_ = GreyscaleExtractor(settings = settings.inverted, **topological_extractor_kwargs)
        self.filtrations_ = FiltrationsExtractor(settings = settings.filtrations, **topological_extractor_kwargs)
        self.point_clouds_ = PointCloudsExtractor(settings = settings.point_clouds, **topological_extractor_kwargs)
        self.geometry_ = GeometryExtractor(settings = settings.geometry, **extractor_kwargs)
        self.scaler_ = sklearn.preprocessing.StandardScaler(copy = False)

    def feature_names(self) -> typing.List[str]:
        return [
            *self.greyscale_.feature_names(),
            *self.inverted_.feature_names(),
            *self.filtrations_.feature_names(),
            *self.point_clouds_.feature_names(),
            *self.geometry_.feature_names()
        ]

    def fit(self, images: numpy.ndarray, dump_name: typing.Optional[str] = None):
        self.fit_transform(images, dump_name)
        return self

    def transform(self, images: numpy.ndarray, dump_name: typing.Optional[str] = None):
        assert self.fitted_ is True, 'fit() must be called before transform()'
        return self.process_(images, do_fit = False, dump_name = dump_name)
    
    def fit_transform(self, images: numpy.ndarray, dump_name: typing.Optional[str] = None):
        result = self.process_(images, do_fit = True, dump_name = dump_name)
        self.fitted_ = True
        return result


    def process_(self, images: numpy.ndarray, do_fit: bool, dump_name: typing.Optional[str] = None):
        results = []

        greyscale_dump = cvtda.dumping.dump_name_concat(dump_name, "greyscale")
        results.append(utils.process_iter(self.greyscale_, images, do_fit, greyscale_dump))
    
        inverted_images = utils.process_iter(self.inverter_, images, do_fit = do_fit)
        inverted_greyscale_dump = cvtda.dumping.dump_name_concat(dump_name, "inverted_greyscale")
        results.append(utils.process_iter(self.inverted_, inverted_images, do_fit, inverted_greyscale_dump))
            
        filtrations_dump = cvtda.dumping.dump_name_concat(dump_name, "filtrations")
        results.append(utils.process_iter(self.filtrations_, images, do_fit, filtrations_dump))
        
        point_clouds_dump = cvtda.dumping.dump_name_concat(dump_name, "point_clouds")
        results.append(utils.process_iter(self.point_clouds_, images, do_fit, point_clouds_dump))

        if not self.return_diagrams_:
            geometry_dump = cvtda.dumping.dump_name_concat(dump_name, "geometry")
            results.append(utils.process_iter(self.geometry_, images, do_fit, geometry_dump))
        
        results = utils.hstack(results, not self.return_diagrams_)
        if self.return_diagrams_:
            cvtda.logging.logger().print("Diagrams requested. Returning diagrams.")
            return results
        
        cvtda.logging.logger().print("Applying StandardScaler.")
        features = utils.process_iter(self.scaler_, results, do_fit)
        if not self.only_get_from_dump_:
            assert features.shape == (len(images), len(self.feature_names()))
        return features



import typing

import numpy
import torch
import pandas
import torchvision

import cvtda.logging
import cvtda.dumping
import cvtda.neural_network

from .Autoencoder import Autoencoder
from .estimate_quality import estimate_quality

def try_autoencoders(
    train_images: numpy.ndarray,
    train_features: numpy.ndarray,
    train_diagrams: typing.Optional[typing.List[numpy.ndarray]],

    test_images: numpy.ndarray,
    test_features: numpy.ndarray,
    test_diagrams: typing.Optional[typing.List[numpy.ndarray]],

    n_jobs: int = -1,
    random_state: int = 42,
    dump_name: typing.Optional[str] = None,
    only_get_from_dump: bool = False,

    nn_device: torch.device = cvtda.neural_network.default_device,
    nn_batch_size: int = 128,
    nn_learning_rate: float = 1e-3,
    nn_epochs: int = 20,
    nn_latent_dim: int = 256,
    nn_base = torchvision.models.resnet34
):
    without_diagrams = (train_diagrams is None) and (test_diagrams is None)

    if (train_images is not None) and (not only_get_from_dump):
        nn_train = cvtda.neural_network.Dataset(
            train_images, train_diagrams, train_features, None, n_jobs = n_jobs, device = nn_device
        )
        nn_test = cvtda.neural_network.Dataset(
            test_images, test_diagrams, test_features, None, n_jobs = n_jobs, device = nn_device
        )

    def try_one(model: Autoencoder, name: str, display_name: str):
        if without_diagrams and name == 'diagrams':
            cvtda.logging.logger().print(f'Skipping {name} - {model}')
            return {}

        cvtda.logging.logger().print(f'Trying {name} - {model}')

        dumper = cvtda.dumping.dumper()
        encoded_dump_name = cvtda.dumping.dump_name_concat(dump_name, f'{name}_encoded')
        decoded_dump_name = cvtda.dumping.dump_name_concat(dump_name, f'{name}_decoded')
        if only_get_from_dump or dumper.has_dump(decoded_dump_name):
            decoded = dumper.get_dump(decoded_dump_name)
        else:
            model.fit(nn_train, nn_test)
            encoded = model.encode(nn_test)
            decoded = model.decode(encoded)

            if encoded_dump_name is not None:
                dumper.save_dump(encoded, encoded_dump_name)
            if decoded_dump_name is not None:
                dumper.save_dump(decoded, decoded_dump_name)
                
        result = { 'model': display_name, **estimate_quality(decoded, test_images) }
        cvtda.logging.logger().print(result)
        return result

    models = [
        Autoencoder(
            random_state = random_state,
            device = nn_device,
            batch_size = nn_batch_size,
            learning_rate = nn_learning_rate * 10,
            n_epochs = nn_epochs * 2,
            latent_dim = nn_latent_dim,
            skip_diagrams = True,
            skip_images = True,
            skip_features = False,
            base = nn_base,
        ),
        Autoencoder(
            random_state = random_state,
            device = nn_device,
            batch_size = nn_batch_size,
            learning_rate = nn_learning_rate,
            n_epochs = nn_epochs,
            latent_dim = nn_latent_dim,
            skip_diagrams = True,
            skip_images = False,
            skip_features = True,
            base = nn_base,
        ),
        Autoencoder(
            random_state = random_state,
            device = nn_device,
            batch_size = nn_batch_size,
            learning_rate = nn_learning_rate,
            n_epochs = nn_epochs,
            latent_dim = nn_latent_dim,
            skip_diagrams = True,
            skip_images = False,
            skip_features = False,
            base = nn_base,
        ),
        Autoencoder(
            random_state = random_state,
            device = nn_device,
            batch_size = nn_batch_size,
            learning_rate = nn_learning_rate,
            n_epochs = nn_epochs // 2,
            latent_dim = nn_latent_dim,
            skip_diagrams = False,
            skip_images = True,
            skip_features = True,
            base = nn_base,
        )
    ]

    names = [
        'features',
        'images',
        'features_images',
        'diagrams'
    ]
    display_names = [
        'FC over topological features',
        'Baseline model',
        'Combined neural network',
        'Trainable vectorization'
    ]
    return pandas.DataFrame([ try_one(*args) for args in zip(models, names, display_names) ])



import torch

def make_conv(in_channels: int, out_channels: int, convolutions: int = 2):
    model = [
        torch.nn.Conv2d(in_channels, out_channels, kernel_size = 3, padding = 1),
        torch.nn.BatchNorm2d(out_channels),
        torch.nn.ReLU(inplace = True)
    ]
    for _ in range(convolutions - 1):
        model.extend([
            torch.nn.Conv2d(out_channels, out_channels, kernel_size = 3, padding = 1),
            torch.nn.BatchNorm2d(out_channels),
            torch.nn.ReLU(inplace = True)
        ])
    return torch.nn.Sequential(*model)



import typing
import dataclasses

import numpy
import gtda.homology

import cvtda.utils
import cvtda.logging

from . import utils
from .interface import TopologicalExtractor
from .DiagramVectorizer import DiagramVectorizer


class PointCloudsExtractor(TopologicalExtractor):
    @dataclasses.dataclass(frozen = True)
    class Settings:
        enabled: bool = True
        vectorizer: DiagramVectorizer.Settings = DiagramVectorizer.Settings()

    PRESETS = cvtda.utils.FeatureExtractorBase.Presets(
        full = Settings(vectorizer = DiagramVectorizer.PRESETS.full),
        reduced = Settings(enabled = False, vectorizer = DiagramVectorizer.PRESETS.reduced),
        quick = Settings(enabled = False, vectorizer = DiagramVectorizer.PRESETS.quick)
    )

    def __init__(
        self,
        n_jobs: int = -1,
        return_diagrams: bool = False,
        settings: Settings = Settings(),
        only_get_from_dump: bool = False,
        **kwargs
    ):
        super().__init__(
            enabled = settings.enabled,
            vectorizer_settings = settings.vectorizer,
            supports_rgb = True,
            n_jobs = n_jobs,
            return_diagrams = return_diagrams,
            settings = settings,
            only_get_from_dump = only_get_from_dump,
            **kwargs
        )

        self.persistence_ = gtda.homology.VietorisRipsPersistence(homology_dimensions = [0, 1, 2], n_jobs = self.n_jobs_)

    def get_diagrams_(self, images: numpy.ndarray, do_fit: bool, dump_name: typing.Optional[str] = None):
        cvtda.logging.logger().print(f"PointCloudsExtractor: processing {dump_name}, do_fit = {do_fit}")
        
        point_clouds = cvtda.utils.image2pointcloud(images, self.n_jobs_)
        return utils.process_iter_dump(self.persistence_, point_clouds, do_fit, self.diagrams_dump_(dump_name))
